/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.resources;

import org.openrdf.repository.RepositoryResult;
import java.util.ArrayList;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.util.Id;
import java.util.List;
import java.util.SortedSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.zip.ZipEntry;
import org.openrdf.model.Statement;
import org.openrdf.model.Value;
import org.openrdf.model.ValueFactory;
import it.tidalwave.bluebill.taxonomy.Taxonomy;
import it.tidalwave.bluebill.taxonomy.TaxonomyManager;
import it.tidalwave.bluebill.taxonomy.elmo.impl.ElmoExporter;
import it.tidalwave.bluebill.taxonomy.io.Exporter;
import it.tidalwave.bluebill.taxonomy.io.Importer;
import it.tidalwave.openrdf.elmo.ElmoManagerThreadLocal;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.logging.Logger;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.zip.ZipOutputStream;
import javax.annotation.Nonnull;
import org.json.me.JSONException;
import org.openrdf.elmo.ElmoManager;
import org.openrdf.elmo.ElmoModule;
import org.openrdf.elmo.sesame.SesameManagerFactory;
import org.openrdf.model.Resource;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryConnection;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.rio.RDFParseException;
import org.openrdf.sail.memory.MemoryStore;
import static it.tidalwave.bluebill.taxonomy.elmo.ElmoTaxonomyVocabulary.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class TaxonomyConverter
  {
    private static final String CLASS = TaxonomyConverter.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    protected final File resourcesFolder;

    protected final File targetFolder;

    protected SesameManagerFactory smf;

    protected Repository repository;

    private ZipOutputStream zos;

    private int count = 0;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static void main (final @Nonnull String ... args)
      throws RepositoryException, IOException, RDFParseException, RDFHandlerException, NotFoundException, JSONException
      {
        final TaxonomyConverter taxonomyConverter = new TaxonomyConverter(args);
        taxonomyConverter.convertTaxonomiesToJson();
        taxonomyConverter.createFachSheetsZip();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private TaxonomyConverter (final @Nonnull String ... args)
      throws IOException
      {
        final String baseDir = args[0];
        resourcesFolder = new File(baseDir + "/target/");
        targetFolder = new File(baseDir + "/target/");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void convertTaxonomiesToJson()
      throws RepositoryException, IOException, RDFParseException, RDFHandlerException, NotFoundException, JSONException
      {
        final Importer importer = Locator.find(TaxonomyManager.class).as(Importer.class);

        initialize();
        importer.import_(repository, new File(targetFolder, "AOU7th.rdf"));
        exportRepository(targetFolder, "AOU7th", "AOU 7th edition");
        close();

        initialize();
        importer.import_(repository, new File(targetFolder, "EBNItalia2003.rdf"));
        exportRepository(targetFolder, "EBNItalia2003", "EBNItalia 2003");
        close();

        initialize();
        importer.import_(repository, new File(targetFolder, "RobertsVII.rdf"));
        exportRepository(targetFolder, "RobertsVII", "Roberts VII");
        close();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void createFachSheetsZip()
      throws RepositoryException, IOException, RDFParseException, RDFHandlerException, NotFoundException, JSONException
      {
        final Importer importer = Locator.find(TaxonomyManager.class).as(Importer.class);
        
        initialize();
        importer.import_(repository, new File(targetFolder, "AOU7th.rdf"));
        importer.import_(repository, new File(targetFolder, "EBNItalia2003.rdf"));
        importer.import_(repository, new File(targetFolder, "RobertsVII.rdf"));
        importer.import_(repository, new File(targetFolder, "ITIS.rdf"));
        importer.import_(repository, new File(targetFolder, "Wikipedia.rdf"));
        importer.import_(repository, new File(targetFolder, "WikipediaPhotos.rdf"));

        final Set<Id> ids = getTaxaIds();
        zos = new ZipOutputStream(new FileOutputStream(new File(targetFolder, "factsheets.zip")));

        for (final Id id : ids)
          {
            createTaxonRepository(id);
          }
        
        zos.close();
        
//        new ElmoExporter(repository).export(new File(targetFolder, "Merged.rdf"));
        close();
      }

    /*******************************************************************************************************************
     *
     * FIXME: should only return species?
     *
     ******************************************************************************************************************/
    @Nonnull
    private SortedSet<Id> getTaxaIds()
      throws RepositoryException
      {
//        return new TreeSet<Id>(Arrays.asList(new Id("urn:lsid:catalogueoflife.org:taxon:eff82234-29c1-102b-9a4a-00304854f820:ac2010")));
        final RepositoryConnection conn = repository.getConnection();
        final ValueFactory valueFactory = conn.getValueFactory();
        final RepositoryResult<Statement> statements = conn.getStatements(null, valueFactory.createURI(URI_SCIENTIFIC_NAME_ID), null, false);
        final SortedSet<Id> ids = new TreeSet<Id>();

        while (statements.hasNext())
          {
            ids.add(new Id(statements.next().getObject().stringValue()));
          }

        conn.close();

        return ids;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void addStatements (final RepositoryConnection conn,
                                  final List<Statement> statements)
      throws RepositoryException
      {
        for (final Statement statement : statements)
          {
            conn.add(statement);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected List<Statement> findStatementsWithSubject (final @Nonnull RepositoryConnection conn,
                                                         final @Nonnull Id taxonId)
      throws RepositoryException
      {
        final ValueFactory valueFactory = conn.getValueFactory();
        return conn.getStatements(valueFactory.createURI(taxonId.stringValue()), null, null, false).asList();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private List<Id> findObjectIdsReferring (final @Nonnull RepositoryConnection conn,
                                             final @Nonnull Id object)
      throws RepositoryException
      {
        final List<Id> result = new ArrayList<Id>();
        final ValueFactory valueFactory = conn.getValueFactory();

        for (final Statement statement : conn.getStatements(null, null, valueFactory.createLiteral(object.stringValue()), false).asList())
          {
            if (!statement.getPredicate().stringValue().equals("http://rs.tdwg.org/dwc/2009-12-07/terms/scientificNameID") &&
                !statement.getPredicate().stringValue().startsWith("http://www.w3.org/2004/02/skos/core#"))
              {
                result.add(new Id(statement.getSubject().stringValue()));
              }
          }

        for (final Statement statement : conn.getStatements(null, null, valueFactory.createURI(object.stringValue()), false).asList())
          {
            if (!statement.getPredicate().stringValue().equals("http://rs.tdwg.org/dwc/2009-12-07/terms/scientificNameID") &&
                !statement.getPredicate().stringValue().startsWith("http://www.w3.org/2004/02/skos/core#"))
              {
                result.add(new Id(statement.getSubject().stringValue()));
              }
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private List<Id> findObjectIdsReferredBy (final @Nonnull RepositoryConnection conn,
                                              final @Nonnull Id subject)
      throws RepositoryException
      {
        final List<Id> result = new ArrayList<Id>();
        final ValueFactory valueFactory = conn.getValueFactory();

        for (final Statement statement : conn.getStatements(valueFactory.createURI(subject.stringValue()), null, null, false).asList())
          {
            final Value object = statement.getObject();

            if (object instanceof Resource)
              {
                result.add(new Id(object.stringValue()));
              }
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void createTaxonRepository (final @Nonnull Id taxonId)
      throws RDFHandlerException, NotFoundException, JSONException, IOException, RepositoryException
      {
        final Repository destRepository = new SailRepository(new MemoryStore());
        destRepository.initialize();
        final RepositoryConnection destConn = destRepository.getConnection();
        final RepositoryConnection srcConn = repository.getConnection();
        
        final Set<Id> objectIds = new TreeSet<Id>();
        objectIds.add(taxonId);
        objectIds.addAll(findObjectIdsReferring(srcConn, taxonId));

        for (final Id id : new ArrayList<Id>(objectIds))
          {
            objectIds.addAll(findObjectIdsReferredBy(srcConn, id));
          }

        for (final Id id : objectIds)
          {
            addStatements(destConn, findStatementsWithSubject(srcConn, id));
          }

        destConn.commit();
        destConn.close();
        srcConn.close();
        
        final String replace = taxonId.stringValue().replace(':', '_');
//        new ElmoExporter(taxonRepository).export(new File(targetFolder, replace + ".rdf"));
//        new ElmoExporter(taxonRepository).export(new File(targetFolder, replace + ".n3"));
        logger.info("Generating %d: %s...", ++count, replace + ".n3");
        zos.putNextEntry(new ZipEntry(replace + ".n3"));
        new ElmoExporter(destRepository).export(zos, replace + ".n3");
        destRepository.shutDown();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void initialize()
      throws RepositoryException
      {
        repository = new SailRepository(new MemoryStore());
        repository.initialize();
        final ElmoModule elmoModule = new ElmoModule();
        smf = new SesameManagerFactory(elmoModule, repository);
        final ElmoManager em = smf.createElmoManager();
        ElmoManagerThreadLocal.set(em);
        targetFolder.mkdirs();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void close()
      {
        ElmoManagerThreadLocal.get().close();
        ElmoManagerThreadLocal.set(null);
        smf.close();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void exportRepository (final @Nonnull File folder,
                                     final @Nonnull String file,
                                     final @Nonnull String taxonomyName)
      throws RepositoryException, IOException, RDFHandlerException, NotFoundException, JSONException
      {
        final Taxonomy taxonomy = Locator.find(TaxonomyManager.class).findTaxonomyByName(taxonomyName, repository);
        taxonomy.as(Exporter.class).export(new File(folder, file + ".n3"));

        final File jsonFile = new File(folder, file + ".json");
        final Writer w3 = new OutputStreamWriter(new FileOutputStream(jsonFile), "UTF-8");
        final TaxonomyJSONExporter exporter = new TaxonomyJSONExporter();
        exporter.export(taxonomy, w3);
        w3.close();
      }
  }